/*-
 * Copyright (c) 2001 Networks Associates Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: kernel_monitor.c,v 1.9 2001/09/26 23:00:44 tfraser Exp $
 */

/*
 * This file contains functions which update LOMAC's internal
 * state in response to system events, such as successful
 * system calls.  These updates allow LOMAC to keep an accurate
 * picture of the kernel's state, enabling LOMAC to make reasonable
 * decisions when it mediates processes' use of security-relevant
 * system calls.  These functions perform no mediation themselves -
 * that is, they do not make access control decisions concerning
 * whether a given system call should be allowed or denied. This
 * mediation is handled by the functions in lomac_mediate.c.
 */

#include "kernel_interface.h"
#include "kernel_monitor.h"
#include "kernel_log.h"

#include "kernel_util.h"


/* monitor_read_object()
 *
 * in:     p_subject - subject that read `p_object'.
 *         p_object  - object read by `p_subject'.
 * out:    nothing
 * return: nothing
 *
 * This function examines the objects read by subjects.  If a subject
 * reads from an object with a level lower than its own, this function
 * reduces the subject's level to match the object's.  This lowering
 * is referred to as "demotion" in much of the LOMAC documentation.
 * This function performs no mediation.
 *
 * This function is for the following kinds of objects:
 *   regular files and FIFOs.
 * It is also used to monitor reads on unnamed pipes.  LOMAC does not
 * consider unnamed pipes to be objects, and treats them differently
 * from objects such as files.  However, the differences are mostly
 * in the write-handling behavior.  LOMAC's read-handling behavior is
 * the same both for objects and unnamed pipes, so this function
 * handles both cases.
 */

int monitor_read_object( lomac_subject_t *p_subject,
			  lomac_object_t *p_object ) {

  lattr_t subject_lattr;     /* lattr of `p_subject' */
  lattr_t object_lattr;      /* lattr of `p_object' */

  /* Get the lattrs of `p_subject' and `p_object' so we can compare them. */
  get_subject_lattr( p_subject, &subject_lattr );
  get_object_lattr(  p_object,  &object_lattr );

  /*
   * If `p_object's level is less than `p_subject's level,
   * we must demote `p_subject'.  The level may be 0 to indicate
   * existence before LOMAC started.
   */
  if ( object_lattr.level && 
       lomac_must_demote( &subject_lattr, &object_lattr ) ) {

    if (subject_do_not_demote(p_subject))
	return (0);
    set_subject_lattr( p_subject, object_lattr );  /* demote! */
    if( lomac_verbose_demote_deny ) {
      lomac_log_t *s = log_start( );
      log_append_string( s, "LOMAC: level-" );
      log_append_int( s, subject_lattr.level );
      log_append_string( s, " subject " );
      log_append_subject_id( s, p_subject );
      log_append_string( s, " demoted to level " );
      log_append_int( s, object_lattr.level );
      log_append_string( s, " after reading " );
      log_append_object_id( s, p_object );
      log_append_string( s, "\n" );
      log_print( s );
    }

#ifdef PARANOID
    get_subject_level( p_subject, &subject_level );
    get_object_level( p_object, &object_level );
    if( !( LEVEL_LEQ( subject_level, object_level ) &&
	   LEVEL_LEQ( object_level, subject_level ) ) ) {
      PANIC( "LOMAC: subject and object levels not equal after demotion\n" );
    }
#endif

  } /* if we need to demote */
  return (0);
} /* monitor_read_object() */


/* monitor_pipe_write()
 *
 * in:     p_subject - subject that just wrote to `p_pipe'.
 *         p_pipe - pipe `p_subject' has written to.
 * out:    p_pipe - pipe may have its level adjusted.
 * return: 0
 *
 *     This function should be called after a successful write to
 * `p_pipe'.  If the level of `p_subject' is less than the level of
 * `p_pipe', this function reduces `p_pipe's level to match
 * `current's.
 *
 *
 */

int 
monitor_pipe_write( lomac_subject_t *p_subject, lomac_object_t *p_pipe ) {

	lattr_t pipe_lattr;          /* lattr of `p_pipe' */
	lattr_t subject_lattr;       /* lattr of `p_subject' */

#ifdef PARANOID
	if( !p_subject ) {
		PANIC( "LOMAC/K: null subject pointer passed to "
		       "monitor_pipe_write().\n" );
	}
	if( !p_pipe ) {
		PANIC( "LOMAC/K: null pipe pointer passed to "
		       "monitor_pipe_write().\n" );
	}
	if( p_pipe->lo_type != LO_TYPE_PIPE ) {
		PANIC( "LOMAC/K: monitor_pipe_write() given non-pipe.\n" );
	}
#endif

	get_subject_lattr( p_subject, &subject_lattr );
	get_object_lattr( p_pipe, &pipe_lattr );
	if( lomac_must_demote( &pipe_lattr, &subject_lattr ) ) {

		subject_lattr.flags = 0;
		set_object_lattr( p_pipe, subject_lattr );
#ifdef LOMAC_DEBUG_PIPE
		if( lomac_verbose_pipe ) {
			lomac_log_t *s = log_start( );
			log_append_string( s, "LOMAC: level-" );
			log_append_int( s, subject_lattr.level );
			log_append_string( s, " subject " );
			log_append_subject_id( s, p_subject );
			log_append_string( s, " contaminated level-" );
			log_append_int( s, pipe_lattr.level );
			log_append_string( s, " pipe " );
			log_append_object_id( s, p_pipe );
			log_append_string( s, "\n" );
			log_print( s );
		}
#endif /* LOMAC_DEBUG_PIPE */
	}

	return ( 0 );

} /* monitor_pipe_write() */


/* monitor_read_net_socket()
 *
 * in:     p_subject - subject that read from network socket.
 * out:    nothing
 * return: 0
 *
 *
 */

int monitor_read_net_socket( lomac_subject_t *p_subject ) {

	lattr_t subject_lattr;          /* lattr of `p_subject' */
	lattr_t socket_lattr;           /* lattr of socket (always lowest) */

	socket_lattr.level = LOMAC_LOWEST_LEVEL;
	socket_lattr.flags = 0;
	get_subject_lattr( p_subject, &subject_lattr );

	if (lomac_must_demote( &subject_lattr, &socket_lattr )) {

		if (subject_do_not_demote(p_subject))
			return (0);
		socket_lattr.flags = subject_lattr.flags;
		set_subject_lattr( p_subject, socket_lattr );  /* demote! */
		if( lomac_verbose_demote_deny ) {
			lomac_log_t *s = log_start();
			log_append_string( s, "LOMAC: level-" );
			log_append_int( s, subject_lattr.level );
			log_append_string( s, " subject " );
			log_append_subject_id( s, p_subject );
			log_append_string( s, " demoted to level " );
			log_append_int( s, socket_lattr.level );
			log_append_string( s, " after reading from "
			    "the network\n" );
			log_print( s );
		}
	}

	return (0);
}

#if 0

/* monitor_open()
 *
 * in:     p_subject - subject that is opening `p_object'.
 *         p_object  - object being opened.
 * out:    nothing
 * return: nothing
 *
 * This function sets the level of objects as they are brought into memory
 * by the open/creat activity of processes.  In this version of LOMAC,
 * the proper level of any object (newly created or not) is determined
 * solely according to the PLM.  This function performs no mediation.
 *
 */

void monitor_open( lomac_subject_t *p_subject, lomac_object_t *p_object ) {

  level_t object_level = LOMAC_INVALID_LEVEL;   /* level of `p_object' */

#ifdef PARANOID
  if( !p_subject ) {
    PANIC( "LOMAC: null subject pointer in monitor_open.\n" );
  }
  if( !p_object ) {
    PANIC( "LOMAC: null object pointer in monitor_open.\n" );
  }
#endif

  /* Optimization - check to see if `p_object' already has its level *
   * set, and if it does, don't bother to set it again.              */
  get_object_level( p_object, &object_level );
  if( object_level != LOMAC_INVALID_LEVEL ) {
    return;
  }


  if( get_object_canabspath( p_object, log_s, LOG_BUFFER_LENGTH ) ) {

#ifdef PARANOID
    /* Error - we are trying to get the level of a nameless object. */
    log_start();
    log_append_string( "LOMAC: unnamed object " );
    log_append_object_id( p_object );
    log_append_string( " in open!\n" );
    log_print();
    PANIC( "LOMAC: open with nameless object, see log\n" );
#endif

  } 

  get_pathname_level( log_s, &object_level );
  set_object_level( p_object, object_level );

  if( verbose & VERBOSE_LOG_OBJECTS ) {
    log_start();
    log_append_string( "LOMAC: " );
    log_append_subject_id( p_subject );
    log_append_string( " opens " );
    log_append_object_id( p_object );
    log_append_string( " level " );
    log_append_int( object_level );
    log_append_string( "\n" );
    log_print();
  }

#ifdef PARANOID
  get_object_level( p_object, &object_level );
  if( object_level == LOMAC_INVALID_LEVEL ) {
    PANIC( "LOMAC: monitor_open left object with invalid level.\n" );
  }
#endif

} /* monitor_open() */




/***********************************************************************
 * The following two routines to allow LOMAC to handle unnamed pipes.
 * LOMAC does not consider unnamed pipes to be objects, and constrains
 * their use as described below:
 * 
 * Originally (v0.3pre3), LOMAC enforced a simple set of constraints
 * that prevented subjects from using unnamed pipes as a conduit for
 * transferring low-intergrity data from low-integrity subjects (jobs)
 * to high-integrity subjects (jobs).  These constraints did not
 * involve assigning levels to unnamed pipes.  LOMAC ensured that a
 * given unnamed pipe was used only by processes in the same job.
 * LOMAC implemented this constraint by marking each pipe with the
 * pgrp number of the first process to access (read or write) it.
 * LOMAC denied subsequent accesses from processes whose pgrp numbers
 * didn't match the first.  This was the scheme described in: Fraser,
 * "LOMAC: Low Water-Mark Integrity Protection for COTS Environments,"
 * 2000 IEEE Symposium on Security and Privacy.
 * 
 * This original scheme worked fine for pipes used in shell pipelines,
 * but caused unacceptable failures in pipes used in some other
 * circumstances, particularly with cron.  (cron/logrotate and
 * cron/procmail were good examples.)  Several children of cron seemed
 * to enjoy sending messages to their parents using unnamed pipes
 * across job boundaries.  Consequently, this original unnamed pipe
 * constraint scheme has been replaced.
 *
 * The current version of LOMAC assigns levels to unnamed pipes.
 * However, unlike the levels it assigns to objects such as files,
 * sockets, and FIFOs, LOMAC allows the levels of unnamed pipes to
 * change over time.  When a subject creates a pipe, the pipe inherits
 * the subject's level (implemented in monitor_pipe_create(), below.)
 * When a low-level subject writes to a high-level pipe, LOMAC reduces
 * the pipe's level to match the level of the writing subject
 * (implemented in monitor_pipe_write(), below.)  When a high-level
 * subject reads from a low-level pipe, LOMAC reduces the reading
 * subject's level to match the pipe's (implemented by the same
 * monitor_read() code used for other types of objects).  LOMAC never
 * prevents a subject from writing to a pipe.
 *
 ***********************************************************************/


/* monitor_pipe_create()
 *
 * in:     p_subject - the subject that created `p_pipe'.
 *         p_pipe    - a newly-created pipe.
 * out:    p_pipe    - the pipe will be marked with the highest level.
 * return: nothing
 *
 *     This function marks `p_pipe' with the the same level as `p_subject'.
 *
 */

void monitor_pipe_create( lomac_subject_t *p_subject, 
			  lomac_object_t *p_pipe ) {

  level_t subject_level;       /* level of `p_subject' */

#ifdef PARANOID
  if( !p_subject ) {
    PANIC( "LOMAC/K: null subject pointer passed to monitor_pipe_create().\n");
  }
  if( !p_pipe ) {
    PANIC( "LOMAC/K: null pipe pointer passed to monitor_pipe_create().\n" );
  }
  if( !IS_PIPE(p_pipe) ) {
    PANIC( "LOMAC/K: monitor_pipe_create() given non-pipe.\n" );
  }
#endif

  get_subject_level( p_subject, &subject_level );
  set_object_level( p_pipe, subject_level );
  if( verbose & VERBOSE_LOG_OBJECTS ) {
    log_start();
    log_append_string( "LOMAC: " );
    log_append_subject_id( p_subject );
    log_append_string( " creates level-" );
    log_append_int( subject_level );
    log_append_string( " pipe " );
    log_append_object_id( p_pipe );
    log_append_string( "\n" );
    log_print();
  }

} /* monitor_pipe_create() */


/* monitor_unix_socket_bind()
 *
 * in:     p_subject - subject doing the BIND operation on `p_socket'.
 *         p_socket  - socket server will use to solicit client connections.
 *         p_name    - `p_socket' is bound to this name in the filesystem.
 * out:    nothing
 * return: nothing
 *
 * LOMAC assigns levels to objects in the filesystem, including filesystem
 * objects that are the names of UNIX (local) domain sockets.  Servers
 * advertize their service by binding a name to their UNIX (local) domain
 * server socket using the BIND operation.  Subsequently, clients may
 * attempt to CONNECT to this server socket.  Since the CONNECT operation
 * operates on the server socket, and not its name in the filesystem, we
 * must ensure that the server socket has the same level as its name in
 * the filesystem.  We do this at BIND time with this function.
 *
 */

void monitor_unix_socket_bind( lomac_subject_t *p_subject,
			       lomac_object_t *p_socket, 
			       lomac_object_t *p_name ) {

  level_t name_level;    /* level of `p_name' */

#ifdef PARANOID
  level_t socket_level;   /* level of `p_socket' */

  if( !p_subject ) {
    PANIC( "LOMAC: null subject pointer in monitor_unix_socket_bind.\n" );
  }
  if( !p_socket ) {
    PANIC( "LOMAC: null socket pointer in monitor_unix_socket_bind.\n" );
  }
  if( !( IS_LOCAL_SOCKET( p_socket ) ) ) {
    PANIC( "LOMAC: monitor_unix_socket_bind given non-UNIX domain socket.\n" );
  }
  if( !p_name ) {
    PANIC( "LOMAC: null new socket pointer in monitor_unix_socket_bind.\n" );
  }

  /* don't try this check - get_object_level prints a warning *
   * when it is used to get LOMAC_INVALID_LEVEL.              *
   * get_object_level( p_socket, &socket_level );             *
   *  if( socket_level != LOMAC_INVALID_LEVEL ) {             */
#endif /* PARANOID */

  /* The BIND operation effectively creates a new object in the *
   * filesystem, specifically `p_name'.  First, we must set     *
   * `p_name's level.                                           */
  monitor_open( p_subject, p_name );

#ifdef PARANOID
  get_object_level( p_name, &name_level );
  if( name_level == LOMAC_INVALID_LEVEL ) {
    PANIC( "LOMAC: name level unset in monitor_unix_socket_bind.\n" );
  }
#endif

  /* Second, we must set p_socket's level to the same level as `p_name'. */
  get_object_level( p_name, &name_level );
  set_object_level( p_socket, name_level );
  
  if( verbose & VERBOSE_LOG_SOCKETS ) {
    log_start();
    log_append_string( "LOMAC: " );
    log_append_subject_id( p_subject );
    log_append_string( " binds " );
    log_append_object_id( p_socket );
    log_append_string( " to " );
    log_append_object_id( p_name );
    log_append_string( " level " );
    log_append_int( name_level );
    log_append_string( "\n" );
    log_print();
  }

#ifdef PARANOID
  get_object_level( p_socket, &socket_level );
  get_object_level( p_name, &name_level );
  if( socket_level != name_level ) {
    PANIC( "LOMAC: socket/name level mismatch in "
	   "monitor_unix_socket_bind.\n" );
  }
  if( socket_level == LOMAC_INVALID_LEVEL ) {
    PANIC( "LOMAC: socket left with invalid  level by "
	   "monitor_unix_socket_bind.\n" );
  }
#endif /* PARANOID */

} /* monitor_unix_socket_bind() */


/* monitor_unix_socket_abstract()
 *
 * in:     p_subject - subject doing  operation on `p_socket'.
 *         p_socket  - socket being operated upon
 * out:    nothing
 * return: nothing
 *
 * Linux 2.2 can bind unix domain sockets to names in an abstract
 * namespace separate from the traditional filesystem namespace.  It's
 * hard for LOMAC to assign levels to these sockets and those other
 * sockets that connect to them, because LOMAC likes to assign levels
 * to dentry structs, and abstract name bindings have no dentry
 * structs.
 *
 * For now, all UNIX-domain sockets bound to abstract names are set to
 * LOMAC_LOWEST_LEVEL.  This isn't a very good solution, because it
 * effectively prevents high-level processes from making use of these
 * sockets.  Sockets that connect to these abstract-bond sockets will
 * also be set to LOMAC_LOWEST_LEVEL.  This function can be used to do
 * all the setting.
 *
 * */

void monitor_unix_socket_abstract( lomac_subject_t *p_subject,
				   lomac_object_t *p_socket ) { 

#ifdef PARANOID
  if( !p_subject ) {
    PANIC( "LOMAC: null subject pointer in "
	   "monitor_unix_socket_bind_abstract.\n" );
  }
  if( !p_socket ) {
    PANIC( "LOMAC: null socket pointer in "
	   "monitor_unix_socket_bind_abstract.\n" );
  }
  if( !( IS_LOCAL_SOCKET( p_socket ) ) ) {
    PANIC( "LOMAC: monitor_unix_socket_bind_abstract "
	   "given non-UNIX domain socket.\n" );
  }

  /* don't try this check - get_object_level prints a warning *
   * when it is used to get LOMAC_INVALID_LEVEL.              *
   * get_object_level( p_socket, &socket_level );             *
   *  if( socket_level != LOMAC_INVALID_LEVEL ) {             */
#endif /* PARANOID */

  set_object_level( p_socket, LOMAC_LOWEST_LEVEL );
  
  if( verbose & VERBOSE_LOG_SOCKETS ) {
    log_start();
    log_append_string( "LOMAC: " );
    log_append_subject_id( p_subject );
    log_append_string( " binds " );
    log_append_object_id( p_socket );
    log_append_string( " to abstract name level " );
    log_append_int( LOMAC_LOWEST_LEVEL );
    log_append_string( "\n" );
    log_print();
  }

} /* monitor_unix_socket_abstract() */


/* monitor_unix_socketpair()
 *
 * in:     p_subject - subject that created UNIX-domain socketpair
 *         p_socket1 - first socket in pair
 *         p_socket2 - second socket in pair
 * out:    nothing
 * return: nothing
 *
 * LOMAC sets the levels of UNIX domain sockets created with
 * socketpair to match the level of the creating process.
 *
 * */

void monitor_unix_socketpair( lomac_subject_t *p_subject,
			      lomac_object_t *p_socket1,
			      lomac_object_t *p_socket2 ) { 

  level_t subject_level;         /* level of `p_subject' */

#ifdef PARANOID
  level_t socket_level;

  if( !p_subject ) {
    PANIC( "LOMAC: null subject pointer in "
	   "monitor_unix_socketpair.\n" );
  }
  if( !( p_socket1 && p_socket2 ) ) {
    PANIC( "LOMAC: null socket pointer in "
	   "monitor_unix_socketpair.\n" );
  }
  if( !( IS_LOCAL_SOCKET( p_socket1 ) &&
	 IS_LOCAL_SOCKET( p_socket2 ) ) ) {
    PANIC( "LOMAC: monitor_unix_socketpair "
	   "given non-UNIX domain socket.\n" );
  }

  /* don't try this check - get_object_level prints a warning *
   * when it is used to get LOMAC_INVALID_LEVEL.              *
   * get_object_level( p_socket, &socket_level );             *
   *  if( socket_level != LOMAC_INVALID_LEVEL ) {             */
#endif /* PARANOID */

  get_subject_level( p_subject, &subject_level );
  set_object_level( p_socket1, subject_level );
  set_object_level( p_socket2, subject_level );
  
  if( verbose & VERBOSE_LOG_SOCKETS ) {
    log_start();
    log_append_string( "LOMAC: " );
    log_append_subject_id( p_subject );
    log_append_string( " creates socketpair " );
    log_append_object_id( p_socket1 );
    log_append_string( " and " );
    log_append_object_id( p_socket2 );
    log_append_string( " at level " );
    log_append_int( subject_level );
    log_append_string( "\n" );
    log_print();
  }

#ifdef PARANOID
  get_object_level( p_socket1, &socket_level );
  if( !( LEVEL_LEQ( subject_level, socket_level ) &&
	 LEVEL_LEQ( socket_level, subject_level ) ) ) {
    PANIC( "LOMAC: monitor_unix_socketpair failed to set socket1 level\n" );
  }
  get_object_level( p_socket2, &socket_level );
  if( !( LEVEL_LEQ( subject_level, socket_level ) &&
	 LEVEL_LEQ( socket_level, subject_level ) ) ) {
    PANIC( "LOMAC: monitor_unix_socketpair failed to set socket2 level\n" );
  }
#endif /* PARANOID */

} /* monitor_unix_socketpair() */


/* monitor_unix_socket_accept_connect()
 *
 * in:    p_subject    - process doing ACCEPT or CONNECT.
 *        p_old_socket - the socket whose level is already correctly set.
 *                       For accept, this is the original socket used to
 *                       receive the incoming connection.  For connect,
 *                       this is the socket created at the server
 *                       by the accept operation.
 *        p_new_socket - the socket whose level needs to be set.  For
 *                       accept, this is the session socket created for the
 *                       connection.  For connect, this is the client's
 *                       socket.
 * out:    nothing
 * return: nothing
 *
 * When a server calls ACCEPT, the kernel creates a new session socket
 * based on the original one used to receive the connection request.
 * We must ensure that the new socket has the same level as the
 * original one.
 *
 * When a client calls CONNECT, the kernel connects its socket
 * with the session socket the server created with ACCEPT.  We must
 * ensure that the client's socket has the same level as the server's
 * session socket.
 *
 */

void monitor_unix_socket_accept_connect( lomac_subject_t *p_subject,
					 lomac_object_t *p_old_socket,
					 lomac_object_t *p_new_socket ) {

  level_t old_socket_level;   /* level of `p_old_socket' */

#ifdef PARANOID
  level_t new_socket_level;   /* level of `p_new_socket' */

  if( !p_subject ) {
    PANIC( "LOMAC: null subject pointer in "
	   "monitor_unix_socket_accept_connect.\n" );
  }
  if( !p_old_socket ) {
    PANIC( "LOMAC: null old socket pointer in "
	   "monitor_unix_socket_accept_connect.\n" );
  }

  /* Don't try the following test... sometimes `p_old_socket' *
   * is a named object in the filesystem, like "/dev/log".    *
   * if( !( IS_LOCAL_SOCKET( p_old_socket ) ) ) {             */

  if( !p_new_socket ) {
    PANIC( "LOMAC: null new socket pointer in "
	   "monitor_unix_socket_accept_connect.\n" );
  }
  if( !( IS_LOCAL_SOCKET( p_new_socket ) ) ) {
    PANIC( "LOMAC: monitor_unix_socket_accept_connect "
	   "given non-local new socket.\n" );
  }
  get_object_level( p_old_socket, &old_socket_level );
  if( old_socket_level == LOMAC_INVALID_LEVEL ) {
    PANIC( "LOMAC: old socket has invalid level in "
	   "monitor_unix_socket_accept_connect.\n" );
  }

  /* Don't try the following test - get_object_level() burps out *
   * a warning if it is used to get a LOMAC_INVALID_LEVEL level. *
   * get_object_level( p_new_socket, &new_socket_level );        *
   * if( new_socket_level != LOMAC_INVALID_LEVEL ) {             */
#endif /* PARANOID */
  
  get_object_level( p_old_socket, &old_socket_level );
  set_object_level( p_new_socket, old_socket_level );

  if( verbose & VERBOSE_LOG_SOCKETS ) {
    log_start();
    log_append_string( "LOMAC: " );
    log_append_subject_id( p_subject );
    log_append_string( " propagates level " );
    log_append_int( old_socket_level );
    log_append_string( " from " );
    log_append_object_id( p_old_socket );
    log_append_string( " to " );
    log_append_object_id( p_new_socket );
    log_append_string( "\n" );
    log_print();
  }

#ifdef PARANOID
  get_object_level( p_old_socket, &old_socket_level );
  get_object_level( p_new_socket, &new_socket_level );
  if( new_socket_level != old_socket_level ) {
    PANIC( "LOMAC: socket level mismatch in "
	   "monitor_unix_socket_accept_connect.\n" );
  }
  if( new_socket_level == LOMAC_INVALID_LEVEL ) {
   PANIC( "LOMAC: socket left with invalid level by "
	  "monitor_unix_socket_accept_connect.\n" );
  } 
#endif /* PARANOID */

} /* handle_unix_domain_accept_connect() */

#endif
