/*
 * LOMAC - Low Water-Mark Mandatory Access Control
 * Copyright (c) 2001 Networks Associates Technology, Inc.
 * Copyright (c) 1982, 1986, 1989, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 *   Redistributions of source code must retain the above copyright
 *   notice, this list of conditions and the following disclaimer.
 * 
 *   Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: kernel_socket.c,v 1.2 2001/10/02 18:20:02 bfeldman Exp $
 *
 * This file implements LOMAC controls over socket operations.  LOMAC
 * gains control of socket operations by interposing on the `struct
 * pr_usrreqs' operations vectors of each `struct protosw'.  This code
 * replaces each `struct pr_usrreqs' with an instance of `struct
 * lomac_pr_usrreqs' containing LOMAC socket control functions.  These
 * socket control functions implement LOMAC's socket controls, and then
 * call the corresponding socket operations from the original `struct
 * pr_usrreqs'.  Each instance of `struct lomac_pr_usrreqs' ends with
 * a pointer to the `struct pr_usrreqs' it replaces.  These pointers
 * allow the LOMAC socket control functions to find their corresponding
 * original `struct pr_usrreqs' functions.
 *
 * This file provides the function lomac_initialize_sockets() to turn
 * socket interposition on.  Once socket iterposition is turned on,
 * the kernel will begin to call LOMAC's socket control functions.
 * This file provides the lomac_uninitialize_sockets() function to
 * turn socket interposition off.  Once socket interposition is turned
 * off, the kernel will no longer call LOMAC's socket control
 * functions.  However, after this call, some processes may still be
 * blocked in LOMAC's socket control functions.  LOMAC's socket
 * control functions use the `global_lomac_ref_count' variable to
 * count the number of processes that are presently in LOMAC's (socket
 * is it safe for the kernel to unload this loadable kernel module.
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/malloc.h>
#include <sys/module.h>
#include <sys/kernel.h>
#include <sys/proc.h>

#include <sys/domain.h>
#include <sys/namei.h>
#include <sys/protosw.h>
#include <sys/socketvar.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/unpcb.h>
#include <sys/uio.h>
#include <sys/vnode.h>

#include "kernel_interface.h"
#include "kernel_socket.h"
#include "kernel_mediate.h"
#include "kernel_monitor.h"
#include "lomacfs.h"

MALLOC_DEFINE(M_LOMAC_USRREQS, "LOMAC-UR", "LOMAC usrreqs");

struct lomac_pr_usrreqs {
	struct pr_usrreqs lomac_pr_usrreqs; /* LOMAC socket control fxns */
	struct pr_usrreqs *orig_pr_usrreqs; /* original socket op vector */
};

int lomac_abort( struct socket * );
int lomac_local_accept( struct socket *, struct sockaddr ** );
int lomac_attach( struct socket *, int, struct thread * );
int lomac_bind( struct socket *, struct sockaddr *, struct thread * );
int lomac_local_bind( struct socket *, struct sockaddr *, struct thread * );
int lomac_local_connect( struct socket *, struct sockaddr *, struct thread * );
int lomac_local_connect2( struct socket *, struct socket * );
int lomac_control( struct socket *, u_long, caddr_t, struct ifnet *,
    struct thread * );
int lomac_local_detach( struct socket * );
int lomac_disconnect( struct socket * );
int lomac_listen( struct socket *, struct thread * );
int lomac_peeraddr( struct socket *, struct sockaddr ** );
int lomac_rcvd( struct socket *, int );
int lomac_rcvoob( struct socket *, struct mbuf *, int );
int lomac_local_send( struct socket *, int, struct mbuf *, struct sockaddr *,
    struct mbuf *, struct thread * );
int lomac_net_send( struct socket *, int, struct mbuf *, struct sockaddr *,
    struct mbuf *, struct thread * );
int lomac_sense( struct socket *, struct stat * );
int lomac_shutdown( struct socket * );
int lomac_sockaddr( struct socket *, struct sockaddr ** );
int lomac_local_sosend( struct socket *, struct sockaddr *, struct uio *, 
    struct mbuf *, struct mbuf *, int, struct thread * );
int lomac_net_sosend( struct socket *, struct sockaddr *, struct uio *, 
    struct mbuf *, struct mbuf *, int, struct thread * );
int lomac_soreceive( struct socket *, struct sockaddr **, struct uio *,
    struct mbuf **, struct mbuf **, int * );
int lomac_local_soreceive( struct socket *, struct sockaddr **, struct uio *,
    struct mbuf **, struct mbuf **, int * );
int lomac_sopoll( struct socket *, int, struct ucred *, struct thread * );

/* This usrreqs structure implements LOMAC's controls on local sockets */
struct pr_usrreqs lomac_local_usrreqs = {
	NULL,
	lomac_local_accept,
	NULL,
	lomac_local_bind,
	lomac_local_connect,
	lomac_local_connect2,
	NULL,
	lomac_local_detach,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	lomac_local_send,
	NULL,
	NULL,
	NULL,
	NULL,
	lomac_local_soreceive,
	NULL
};

/* This usrreqs structure implements LOMAC's controls on network sockets */
struct pr_usrreqs lomac_net_usrreqs = {
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	lomac_rcvoob,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL,
	lomac_soreceive,
	NULL
};

static __inline struct pr_usrreqs *
orig_pr_usrreqs( struct socket *so ) {
	return (((struct lomac_pr_usrreqs *)(so->so_proto->pr_usrreqs))->
	    orig_pr_usrreqs);
}

int
lomac_abort( struct socket *so ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_abort)( so );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_local_accept( struct socket *so, struct sockaddr **nam ) {
	struct vnode *vp;
	struct unpcb *unp;
	int ret_val;     /* value to return to caller */
	
	unp = sotounpcb(so);
	if (unp == NULL)
		return (EINVAL);
	global_lomac_ref_count++;
	if (unp->unp_conn != NULL) {
		vp = unp->unp_vnode = unp->unp_conn->unp_vnode;
		if (vp != NULL)
			vref(vp);
	}
	ret_val = (*orig_pr_usrreqs( so )->pru_accept)( so, nam );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_attach( struct socket *so, int proto, struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_attach)( so, proto, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_bind( struct socket *so, struct sockaddr *nam, struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_bind)( so, nam, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_local_bind( struct socket *so, struct sockaddr *nam, struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_bind)( so, nam, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_local_connect(struct socket *so, struct sockaddr *nam, struct thread *td)
{
	register struct sockaddr_un *soun = (struct sockaddr_un *)nam;
	register struct vnode *vp;
	register struct socket *so2, *so3;
	struct unpcb *unp, *unp2, *unp3;
	int error, len;
	struct nameidata nd;
	char buf[SOCK_MAXADDRLEN];

	len = nam->sa_len - offsetof(struct sockaddr_un, sun_path);
	if (len <= 0)
		return EINVAL;
	strncpy(buf, soun->sun_path, len);
	buf[len] = 0;

	NDINIT(&nd, LOOKUP, FOLLOW | LOCKLEAF, UIO_SYSSPACE, buf, td);
	global_lomac_ref_count++;
	error = namei(&nd);
	if (error)
		goto bad2;
	vp = nd.ni_vp;
	NDFREE(&nd, NDF_ONLY_PNBUF);
	if (vp->v_type != VSOCK) {
		error = ENOTSOCK;
		goto bad;
	}
	error = VOP_ACCESS(vp, VWRITE, td->td_proc->p_ucred, td);
	if (error)
		goto bad;
	so2 = vp->v_socket;
	if (so2 == 0) {
		error = ECONNREFUSED;
		goto bad;
	}
	if (so->so_type != so2->so_type) {
		error = EPROTOTYPE;
		goto bad;
	}
	if (so->so_proto->pr_flags & PR_CONNREQUIRED) {
		if ((so2->so_options & SO_ACCEPTCONN) == 0 ||
		    (so3 = sonewconn3(so2, 0, td)) == 0) {
			error = ECONNREFUSED;
			goto bad;
		}
		unp = sotounpcb(so);
		unp2 = sotounpcb(so2);
		unp3 = sotounpcb(so3);
		if (unp2->unp_addr)
			unp3->unp_addr = (struct sockaddr_un *)
				dup_sockaddr((struct sockaddr *)
					     unp2->unp_addr, 1);

		/*
		 * unp_peercred management:
		 *
		 * The connecter's (client's) credentials are copied
		 * from its process structure at the time of connect()
		 * (which is now).
		 */
		memset(&unp3->unp_peercred, '\0', sizeof(unp3->unp_peercred));
		unp3->unp_peercred.cr_uid = td->td_proc->p_ucred->cr_uid;
		unp3->unp_peercred.cr_ngroups = td->td_proc->p_ucred->cr_ngroups;
		memcpy(unp3->unp_peercred.cr_groups, td->td_proc->p_ucred->cr_groups,
		    sizeof(unp3->unp_peercred.cr_groups));
		unp3->unp_flags |= UNP_HAVEPC;
		/*
		 * The receiver's (server's) credentials are copied
		 * from the unp_peercred member of socket on which the
		 * former called listen(); unp_listen() cached that
		 * process's credentials at that time so we can use
		 * them now.
		 */
		KASSERT(unp2->unp_flags & UNP_HAVEPCCACHED,
		    ("unp_connect: listener without cached peercred"));
		memcpy(&unp->unp_peercred, &unp2->unp_peercred,
		    sizeof(unp->unp_peercred));
		unp->unp_flags |= UNP_HAVEPC;

		so2 = so3;
	}
	error = lomac_local_connect2(so, so2);
bad:
	vput(vp);
bad2:
	global_lomac_ref_count--;
	return (error);
}

int
lomac_local_connect2( struct socket *so1, struct socket *so2 ) {
	struct vnode *vp;
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	if (so2->so_head != NULL) {
		vp = sotounpcb(so2->so_head)->unp_vnode;
		if (vp != NULL) {
			sotounpcb(so1)->unp_vnode = vp;
			vref(vp);
		}
	}
	ret_val = (*orig_pr_usrreqs( so1 )->pru_connect2)( so1, so2 );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_control( struct socket *so, u_long cmd, caddr_t data,
    struct ifnet *ifp, struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_control)( so, cmd, data,
            ifp, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_local_detach( struct socket *so ) {
	int ret_val;     /* value to return to caller */
	struct unpcb *unp = sotounpcb(so);
	
	if (unp == NULL)
		return (EINVAL);
	global_lomac_ref_count++;
	if (unp->unp_vnode != NULL && unp->unp_vnode->v_socket != so) {
		vrele(unp->unp_vnode);
		unp->unp_vnode = NULL;
	}
	ret_val = (*orig_pr_usrreqs( so )->pru_detach)( so );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_disconnect( struct socket *so ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_disconnect)( so );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_listen( struct socket *so, struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_listen)( so, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_peeraddr( struct socket *so, struct sockaddr **nam ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_peeraddr)( so, nam );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_rcvd( struct socket *so, int flags ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_rcvd)( so, flags );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_rcvoob( struct socket *so, struct mbuf *m, int flags ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_rcvoob)( so, m, flags );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_local_send( struct socket *so, int flags, struct mbuf *m, 
    struct sockaddr *addr, struct mbuf *control, struct thread *td ) {
	struct vnode *vp;
	struct unpcb *unp = sotounpcb(so);
	int error;

	global_lomac_ref_count++;
	/* printf( "pid %d local send\n", p->p_pid ); */
	if (unp == NULL) {
		error = EINVAL;
		goto out;
	}
	if (so->so_type == SOCK_DGRAM) {
		if (addr != NULL) {
			if (unp->unp_conn != NULL) {
				error = EISCONN;
				goto out;
			}
			error = lomac_local_connect(so, addr, td);
			if (error)
				goto out;
		} else if (unp->unp_conn == NULL) {
			error = ENOTCONN;
			goto out;
		}
	} else if ((so->so_state & SS_ISCONNECTED) == 0) {
		if (addr != NULL) {
			error = lomac_local_connect(so, addr, td);
			if (error)
				goto out;	/* XXX */
		} else {
			error = ENOTCONN;
			goto out;
		}
	}
	vp = unp->unp_vnode;
	if (vp != NULL) {
		lomac_object_t lobj;

		lobj.lo_type = VISLOMAC(vp) ? LO_TYPE_LVNODE : LO_TYPE_UVNODE;
		lobj.lo_object.vnode = vp;
		if (!mediate_subject_object("send", td->td_proc, &lobj)) {
			error = EPERM;
			goto out;
		}
	}
	error = (*orig_pr_usrreqs( so )->pru_send)( so, flags, m, NULL,
            control, td );
	if (addr != NULL && so->so_type == SOCK_DGRAM)
		(*orig_pr_usrreqs( so )->pru_disconnect)( so );
out:
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (error);
}

int
lomac_net_send( struct socket *so, int flags, struct mbuf *m, 
    struct sockaddr *addr, struct mbuf *control, struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	/* printf( "pid %d net send\n", p->p_pid ); */
	ret_val = (*orig_pr_usrreqs( so )->pru_send)( so, flags, m, addr,
            control, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}


int
lomac_sense( struct socket *so, struct stat *sb ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_sense)( so, sb );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_shutdown( struct socket *so ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_shutdown)( so );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_sockaddr( struct socket *so, struct sockaddr **nam ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_sockaddr)( so, nam );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_local_sosend( struct socket *so, struct sockaddr *addr, struct uio *uio,
    struct mbuf *top, struct mbuf *control, int flags, struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	/* printf( "pid %d local sosend\n", p->p_pid ); */
	ret_val = (*orig_pr_usrreqs( so )->pru_sosend)( so, addr, uio, top,
            control, flags, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_net_sosend( struct socket *so, struct sockaddr *addr, struct uio *uio,
    struct mbuf *top, struct mbuf *control, int flags, struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	/* printf( "pid %d net sosend\n", p->p_pid ); */
	ret_val = (*orig_pr_usrreqs( so )->pru_sosend)( so, addr, uio, top,
            control, flags, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_local_soreceive( struct socket *so, struct sockaddr **paddr,
    struct uio *uio, struct mbuf **mp0, struct mbuf **controlp, int *flagsp ) {
	lomac_object_t lobj;
	struct vnode *vp;
	struct unpcb *unp = sotounpcb(so);
	int ret_val;     /* value to return to caller */

	if (unp == NULL)
		return (EINVAL);
	vp = unp->unp_vnode;
	global_lomac_ref_count++;
	if (vp != NULL) {
		/*
		 * XXX - ignoring socketpairs.  This is wrong, bad and evil.
		 * Really.  I mean it.
		 */
		lobj.lo_type = VISLOMAC(vp) ? LO_TYPE_LVNODE : LO_TYPE_UVNODE;
		lobj.lo_object.vnode = vp;
		ret_val = monitor_read_object( uio->uio_td->td_proc, &lobj );
		if (ret_val)
			goto out;
	}
	ret_val = (*orig_pr_usrreqs( so )->pru_soreceive)( so, paddr, uio, mp0,
            controlp, flagsp );
out:
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_soreceive( struct socket *so, struct sockaddr **paddr, struct uio *uio,
    struct mbuf **mp0, struct mbuf **controlp, int *flagsp ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_soreceive)( so, paddr, uio, mp0,
            controlp, flagsp );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}

int
lomac_sopoll( struct socket *so, int events, struct ucred *cred,
    struct thread *td ) {
	int ret_val;     /* value to return to caller */
	
	global_lomac_ref_count++;
	ret_val = (*orig_pr_usrreqs( so )->pru_sopoll)( so, events, cred, td );
	global_lomac_ref_count--;
	wakeup(&global_lomac_ref_count);
	return (ret_val);
}


int
lomac_initialize_sockets( void ) {
	struct domain *dp;         /* used to traverse global `domains' list */
	struct protosw *pr;   /* used to traverse each domain's protosw list */
	struct lomac_pr_usrreqs *lomac_pr_usrreqs;  /* lomac usrreqs vectors */
	int nreq = sizeof(struct pr_usrreqs) / sizeof(void (*)(void));
	void (**lfuncp)(void), (**funcp)(void);

	for (dp = domains; dp; dp = dp->dom_next) {

		for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; 
		    pr++) {
			int n;

			lomac_pr_usrreqs = (struct lomac_pr_usrreqs *)
			    malloc( sizeof( struct lomac_pr_usrreqs ),
			    M_LOMAC_USRREQS, M_WAITOK );

			if (lomac_pr_usrreqs == NULL)
				return ENOMEM;

			if (dp->dom_family == AF_LOCAL) {
				memcpy( lomac_pr_usrreqs, 
				    &lomac_local_usrreqs, 
				    sizeof( struct pr_usrreqs ) );
			} else {
				memcpy( lomac_pr_usrreqs, 
				    &lomac_net_usrreqs, 
				    sizeof( struct pr_usrreqs ) );
			}
			/*
			 * Do sparse allocation of user requests and only
			 * override the ones we need to (to reduce overhead).
			 */
			lfuncp = (void (**)(void))lomac_pr_usrreqs;
			funcp = (void (**)(void))pr->pr_usrreqs;
			for (n = 0; n < nreq; n++) {
				if (lfuncp[n] == NULL)
					lfuncp[n] = funcp[n];
			}

			lomac_pr_usrreqs->orig_pr_usrreqs = pr->pr_usrreqs;
			pr->pr_usrreqs = (struct pr_usrreqs *)lomac_pr_usrreqs;
			
		}

	}
	
	return ( 0 );
}


int
lomac_uninitialize_sockets( void ) {

	struct domain *dp;         /* used to traverse global `domains' list */
	struct protosw *pr;   /* used to traverse each domain's protosw list */
	struct lomac_pr_usrreqs *lomac_pr_usrreqs;  /* lomac usrreqs vectors */

	for (dp = domains; dp; dp = dp->dom_next) {

		for (pr = dp->dom_protosw; pr < dp->dom_protoswNPROTOSW; pr++) {

			lomac_pr_usrreqs = (struct lomac_pr_usrreqs *)
			    pr->pr_usrreqs;
			pr->pr_usrreqs = lomac_pr_usrreqs->orig_pr_usrreqs;
			while (global_lomac_ref_count != 0)
				tsleep(&global_lomac_ref_count, PUSER,
				    "lmuisk", 0);
			free( lomac_pr_usrreqs, M_LOMAC_USRREQS );

		}

	}
	
	return ( 0 );
}
