/*-
 * Copyright (c) 2001 Networks Associates Technology, Inc.
 * All rights reserved.
 * Copyright (c) 1982, 1986, 1989, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 * (c) UNIX System Laboratories, Inc.
 * All or some portions of this file are derived from material licensed
 * to the University of California by American Telephone and Telegraph
 * Co. or Unix System Laboratories, Inc. and are reproduced herein with
 * the permission of UNIX System Laboratories, Inc.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: kernel_util.c,v 1.14 2001/10/02 18:20:03 bfeldman Exp $
 */

#include <sys/param.h>
#include <sys/proc.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/signalvar.h>
#include <sys/sx.h>
#include <sys/syscall.h>
#include <sys/sysent.h>
#include <sys/sysproto.h>
#include <sys/systm.h>
#include <sys/linker.h>
#include <sys/unistd.h>

#include "kernel_interface.h"
#include "kernel_util.h"
#include "kernel_mediate.h"
#include "kernel_monitor.h"
#include "lomacfs.h"

#include "syscall_gate/syscall_gate.h"

#define	AS(name) (sizeof(struct name) / sizeof(register_t))

int
each_proc(int (*iter)(struct proc *p)) {
	struct proc *p;
	int error = 0;

	sx_slock(&allproc_lock);
	LIST_FOREACH(p, &allproc, p_list) {
		PROC_LOCK(p);
		error = (*iter)(p);
		PROC_UNLOCK(p);
		if (error)
			goto out;
	}
out:
	sx_sunlock(&allproc_lock);
	return (error);
}

static int
initialize_proc(struct proc *p) {
	lattr_t lattr = { LOMAC_HIGHEST_LEVEL, 0 };

	init_subject_lattr(p, &lattr);
	return (0);
}

static void
lomac_at_fork(struct proc *parent, struct proc *p, int flags) {

	if ((flags & RFMEM) == 0) {
		lattr_t parent_lattr;

		get_subject_lattr(parent, &parent_lattr);
		init_subject_lattr(p, &parent_lattr);
	}
}

int
lomac_initialize_procs(void) {
	int error;

	error = at_fork(lomac_at_fork);
	if (error)
		return (error);
	return (each_proc(&initialize_proc));
}

int
lomac_uninitialize_procs(void) {

	rm_at_fork(lomac_at_fork);
	return (0);
}

/*
 * Common code for kill process group/broadcast kill.
 * cp is calling process.
 */
int
lomac_killpg1(struct proc *cp, int sig, int pgid, int all) {
	struct proc *p;
	struct pgrp *pgrp;
	int nfound = 0;
	lattr_t lattr;

	get_subject_lattr(cp, &lattr);
	if (all) {
		/*
		 * broadcast
		 */
		sx_slock(&allproc_lock);
		LIST_FOREACH(p, &allproc, p_list) {
			PROC_LOCK(p);
			if (p->p_pid <= 1 || p->p_flag & P_SYSTEM || p == cp) {
				PROC_UNLOCK(p);
				continue;
			}
			if (p_cansignal(cp, p, sig) == 0 &&
			    mediate_subject_level_subject("kill", cp,
			    lattr.level, p)) {
				nfound++;
				if (sig)
					psignal(p, sig);
			}
			PROC_UNLOCK(p);
		}
		sx_sunlock(&allproc_lock);
	} else {
		if (pgid == 0)
			/*
			 * zero pgid means send to my process group.
			 */
			pgrp = cp->p_pgrp;
		else {
			pgrp = pgfind(pgid);
			if (pgrp == NULL)
				return (ESRCH);
		}
		LIST_FOREACH(p, &pgrp->pg_members, p_pglist) {
			PROC_LOCK(p);	      
			if (p->p_pid <= 1 || p->p_flag & P_SYSTEM) {
				PROC_UNLOCK(p);
				continue;
			}
			mtx_lock_spin(&sched_lock);
			if (p->p_stat == SZOMB) {
				mtx_unlock_spin(&sched_lock);
				PROC_UNLOCK(p);
				continue;
			}
			mtx_unlock_spin(&sched_lock);
			if (p_cansignal(cp, p, sig) == 0 &&
			    mediate_subject_level_subject("kill", cp,
			    lattr.level, p)) {
				nfound++;
				if (sig)
					psignal(p, sig);
			}
			PROC_UNLOCK(p);
		}
	}
	return (nfound ? 0 : ESRCH);
}

#ifndef _SYS_SYSPROTO_H_
struct kill_args {
	int pid;
	int signum;
};
#endif
/* ARGSUSED */
int
kill(struct thread *td, struct kill_args *uap) {
	struct proc *cp, *p;

	cp = td->td_proc;
	if ((u_int)uap->signum > _SIG_MAXSIG)
		return (EINVAL);
	if (uap->pid > 0) {
		lattr_t lattr;

		/*
		 * This is a race, but it really doesn't
		 * matter at all in the signal case.
		 */
		get_subject_lattr(cp, &lattr);	
		/* kill single process */
		if ((p = pfind(uap->pid)) == NULL)
			return (ESRCH);
		/*
		 * Allow sending signals to init(8) so that the
		 * system can actually be rebooted if in fact that
		 * becomes necessary.
		 */
		if (p_cansignal(cp, p, uap->signum) || (uap->pid != 1 &&
		    !mediate_subject_level_subject("kill", cp, lattr.level,
		    p))) {
			PROC_UNLOCK(p);
			return (EPERM);
		}
		if (uap->signum)
			psignal(p, uap->signum);
		PROC_UNLOCK(p);
		return (0);
	}
	switch (uap->pid) {
	case -1:		/* broadcast signal */
		return (lomac_killpg1(cp, uap->signum, 0, 1));
	case 0:			/* signal own process group */
		return (lomac_killpg1(cp, uap->signum, 0, 0));
	default:		/* negative explicit process group */
		return (lomac_killpg1(cp, uap->signum, -uap->pid, 0));
	}
	/* NOTREACHED */
}

extern int (*old_execve)(struct thread *, struct execve_args *);

int
execve(struct thread *td, struct execve_args *uap) {
	lattr_t lattr;
	struct vmspace *oldvmspace;
	struct proc *p;
	int error;

	p = td->td_proc;
	get_subject_lattr(p, &lattr);
	oldvmspace = p->p_vmspace;
	error = old_execve(td, uap);
	if (error == 0) {
		lomac_object_t lobj;

		if (p->p_vmspace != oldvmspace)
			init_subject_lattr(p, &lattr);
		lobj.lo_type = VISLOMAC(p->p_textvp) ? LO_TYPE_LVNODE :
		    LO_TYPE_UVNODE;
		lobj.lo_object.vnode = p->p_textvp;
		mtx_lock(&Giant);
		(void)monitor_read_object(p, &lobj);
		mtx_unlock(&Giant);
	}
	return (error);
}

const char *linker_basename(const char* path);
int linker_load_module(const char *kldname, const char *modname,
	struct linker_file *parent, struct mod_depend *verinfo,
	struct linker_file **lfpp);

MALLOC_DECLARE(M_LINKER);

/*
 * MPSAFE
 */
int
kldload(struct thread* td, struct kldload_args* uap)
{
    char *kldname, *modname;
    char *pathname = NULL;
    linker_file_t lf;
    int error = 0;

    td->td_retval[0] = -1;

    if (securelevel > 0)	/* redundant, but that's OK */
	return EPERM;

    mtx_lock(&Giant);

    if ((error = suser_td(td)) != 0)
	goto out;

    pathname = malloc(MAXPATHLEN, M_TEMP, M_WAITOK);
    if ((error = copyinstr(SCARG(uap, file), pathname, MAXPATHLEN, NULL)) != 0)
	goto out;
    if (!mediate_subject_at_level("kldload", td->td_proc,
	LOMAC_HIGHEST_LEVEL)) {
	error = EPERM;
	goto out;
    }

    /*
     * If path do not contain qualified name or any dot in it (kldname.ko, or
     * kldname.ver.ko) treat it as interface name.
     */
    if (index(pathname, '/') || index(pathname, '.')) {
	kldname = pathname;
	modname = NULL;
    } else {
	kldname = NULL;
	modname = pathname;
    }
    error = linker_load_module(kldname, modname, NULL, NULL, &lf);
    if (error)
	goto out;

    lf->userrefs++;
    td->td_retval[0] = lf->id;

out:
    if (pathname)
	free(pathname, M_TEMP);
    mtx_unlock(&Giant);
    return (error);
}


#ifdef __i386__
#include <machine/sysarch.h>

extern int (*old_sysarch)(struct thread *, void *);

int
sysarch(struct thread *td, struct sysarch_args *uap) {
	switch (uap->op) {
	case I386_SET_IOPERM:
		if (!mediate_subject_at_level("ioperm", td->td_proc,
		    LOMAC_HIGHEST_LEVEL))
			return (EPERM);
	default:
		return (old_sysarch(td, uap));
	}
}
#endif

extern int lomac_mmap(struct proc *, struct mmap_args *);

static struct syscall_override {
	int offset;
	sy_call_t *call;
	int narg;
	int mpsafe;
} syscall_overrides[] = {
	{ SYS_kill, (sy_call_t *)kill, AS(kill_args), 1 },
	{ SYS_mmap, (sy_call_t *)mmap, AS(mmap_args), 1 },
	{ SYS_execve, (sy_call_t *)execve, AS(execve_args), 1 },
	{ SYS_kldload, (sy_call_t *)kldload, AS(kldload_args), 1 },
#ifdef __i386__
	{ SYS_sysarch, (sy_call_t *)sysarch, AS(sysarch_args), 1 }
#endif
};

int
lomac_initialize_syscalls(void) {
	int error, i;

	for (i = 0; i < sizeof(syscall_overrides) / sizeof(syscall_overrides[0]); i++) {
		struct syscall_override *so = &syscall_overrides[i];

		error = syscall_gate_register(so->offset, so->call, so->narg, so->mpsafe);
		if (error) {
			while (--i >= 0)
				syscall_gate_deregister(syscall_overrides[i].offset);
			return (error);
		}
	}
	return (0);
}

int
lomac_uninitialize_syscalls(void) {
	int i;

	for (i = 0; i < sizeof(syscall_overrides) / sizeof(syscall_overrides[0]); i++)
		syscall_gate_deregister(syscall_overrides[i].offset);
	return (0);
}
