/*-
 * Copyright (c) 2001 Networks Associates Technology, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: syscall_gate.c,v 1.3 2001/10/02 18:20:03 bfeldman Exp $
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/module.h>
#include <sys/lock.h>
#include <sys/mutex.h>
#include <sys/proc.h>
#include <sys/syscall.h>
#include <sys/sysent.h>

#include <machine/frame.h>

#include "syscall_gate.h"
void syscall_gate_init(void);
int syscall_gate(struct thread *td, caddr_t params);

static struct syscall_gate sg;

void
syscall_gate_init(void) {
	int i;

	sx_init(&sg.sg_lock, "scgate");
	sg.sg_table = curthread->td_proc->p_sysent->sv_table;
	bzero(sg.sg_oldsyscalls, sizeof(sg.sg_oldsyscalls));
	for (i = 0; i < SYS_MAXSYSCALL; i++)
		sg.sg_newsyscalls[i] = sg.sg_table[i];
}

int
syscall_gate(struct thread *td, caddr_t params) {
	struct trapframe *framep = td->td_frame;
	struct sysent *callp;
	int error, narg;
	int hidden = 0;
#ifdef __i386__
	u_int code = framep->tf_eax;
	int args[8];
	params = (caddr_t)framep->tf_esp + sizeof(int);
#elif defined(__alpha__)
	u_int64_t code = framep->tf_regs[FRAME_V0];
	u_int64_t args[10];					/* XXX */                   
#else
#error "unknown FreeBSD platform"
#endif

#ifdef __i386__
	if (code == SYS_syscall) {
		/*
		 * Code is first argument, followed by actual args.
		 */
		code = fuword(params);
		params += sizeof(int);
	} else if (code == SYS___syscall) {
		/*
		 * Like syscall, but code is a quad, so as to maintain
		 * quad alignment for the rest of the arguments.
		 */
		code = fuword(params);
		params += sizeof(quad_t);
	}
#elif defined(__alpha__)
	/*
	 * syscall() and __syscall() are handled the same on
	 * the alpha, as everything is 64-bit aligned, anyway.
	 */
	if (code == SYS_syscall || code == SYS___syscall) {
		/*
		 * Code is first argument, followed by actual args.
		 */
		code = framep->tf_regs[FRAME_A0];
		hidden = 1;
	}
#endif
	if (curthread->td_proc->p_sysent->sv_mask)
		code &= curthread->td_proc->p_sysent->sv_mask;

	sx_slock(&sg.sg_lock);
	if (code >= SYS_MAXSYSCALL)
		callp = &sg.sg_newsyscalls[0];
	else
		callp = &sg.sg_newsyscalls[code];

	narg = (callp->sy_narg & SYF_ARGMASK) + hidden;
#ifdef __i386__
	if (narg != 0 && copyin(params, (caddr_t)args, (u_int)narg * sizeof(int)) != 0) {
		sx_sunlock(&sg.sg_lock);
		return (EFAULT);
	}
#elif defined(__alpha__)
	switch (nargs) {
	default:
		if (nargs > 10)		/* XXX */
			panic("syscall: too many args (%d)", nargs);
		if (copyin((caddr_t)(alpha_pal_rdusp()), &args[6],
		    (nargs - 6) * sizeof(u_int64_t)) != 0) {
			sx_sunlock(&sg.sg_lock);
			return (EFAULT);
		}
	case 6:	
		args[5] = framep->tf_regs[FRAME_A5];
	case 5:	
		args[4] = framep->tf_regs[FRAME_A4];
	case 4:	
		args[3] = framep->tf_regs[FRAME_A3];
	case 3:	
		args[2] = framep->tf_regs[FRAME_A2];
	case 2:	
		args[1] = framep->tf_regs[FRAME_A1];
	case 1:	
		args[0] = framep->tf_regs[FRAME_A0];
	case 0:
		break;
	}
#endif
	/*
	 * Try to run the syscall without the MP lock if the syscall
	 * is MP safe
	 */
	if ((callp->sy_narg & SYF_MPSAFE) == 0) {
		mtx_lock(&Giant);
	}
	error = callp->sy_call(td, args + hidden);
	/* syscall() itself will release Giant */
	sx_sunlock(&sg.sg_lock);
	return (error);
}

int
syscall_gate_register(int offset, sy_call_t *call, int narg, int mpsafe) {
	struct sysent *se;
	int error = 0;

	sx_xlock(&sg.sg_lock);
	if (offset <= 0 || offset == SYS_syscall || offset == SYS___syscall ||
	    offset >= SYS_MAXSYSCALL) {
		error = EINVAL;
		goto out;
	}
	if (sg.sg_oldsyscalls[offset].sy_call != NULL) {
		error = EEXIST;
		goto out;
	}
	se = &sg.sg_table[offset];
	sg.sg_oldsyscalls[offset] = *se;
	se->sy_call = (sy_call_t *)syscall_gate;
	se->sy_narg = SYF_MPSAFE;
	sg.sg_newsyscalls[offset].sy_call = call;
	sg.sg_newsyscalls[offset].sy_narg = narg | (mpsafe ? SYF_MPSAFE : 0);
out:
	sx_xunlock(&sg.sg_lock);
	return (error);
}

void
syscall_gate_deregister(int offset) {
	KASSERT(offset > 0 && offset < SYS_MAXSYSCALL, ("syscall offset %d out of range",
	    offset));
	sx_xlock(&sg.sg_lock);
	KASSERT(sg.sg_oldsyscalls[offset].sy_call != NULL, ("deregistering nonexistant syscall %d",
	    offset));
	sg.sg_table[offset] = sg.sg_oldsyscalls[offset];
	sg.sg_oldsyscalls[offset].sy_call = NULL;
	sg.sg_oldsyscalls[offset].sy_narg = 0;
	sx_xunlock(&sg.sg_lock);
}

static int
syscall_gate_modevent(module_t module, int event, void *unused) {
	int i;

	switch ((enum modeventtype)event) {
	case MOD_LOAD:
		syscall_gate_init();
		break;
	case MOD_UNLOAD:
		sx_xlock(&sg.sg_lock);
		for (i = 1; i < SYS_MAXSYSCALL; i++) {
			struct sysent *se;

			se = &sg.sg_oldsyscalls[i];
			if (se->sy_call != NULL)
				sg.sg_table[i] = *se;
		}
		sx_xunlock(&sg.sg_lock);
		sx_destroy(&sg.sg_lock);
		break;
	case MOD_SHUTDOWN:
		break;
	}
	return (0);
}

static moduledata_t syscall_gate_moduledata = {
	"syscall_gate",
	&syscall_gate_modevent,
	NULL
};
DECLARE_MODULE(syscall_gate, syscall_gate_moduledata, SI_SUB_DRIVERS, SI_ORDER_ANY);
MODULE_VERSION(syscall_gate, 1);
